import os, glob
from pdf2image import convert_from_path

class PDFConverter:
    def __init__(self, pdfFile, imageDir):
        self.pdfFile = pdfFile
        self.imageDir = imageDir

    def generateImage(pdfFile, imageDir):
        imgPaths = []
        pdfImgPath = os.path.join(imageDir, os.path.basename(pdfFile).replace('.pdf', '').replace('.PDF', ''))
        
        if not os.path.exists(pdfImgPath):
            os.makedirs(pdfImgPath)

        pdfImages = convert_from_path(pdfFile, dpi=500)
        imgName = os.path.basename(pdfFile).replace('.pdf', '').replace('.PDF', '')

        for i, img in enumerate(pdfImages):
            imgPath = os.path.join(pdfImgPath, f'{imgName}_page{i}.jpg')
            img.save(imgPath, 'JPEG')
            imgPaths.append(imgPath)

        return imgPaths


import os, shutil
import cv2
import numpy as np
import easyocr
import pandas as pd
import concurrent.futures
import warnings
warnings.simplefilter(action='ignore', category=UserWarning)

from .convertor import PDFConverter
from .rotate import ImageRotator
from .bordered import BorderedTableDetector
from .unbordered import UnborderedTableDetector
from .semibordered import SemiborderedTableDetector
from .consolidated import ConsolidatedTableDetector
from .extractor import TableExtractor



class TableDetector:
    def __init__(self, pdfPath, method="consolidated"):
        self.pdfPath = pdfPath
        self.method = method

    def processImage(self, eachImgPath, ocrEngine):

        try:
            rotator = ImageRotator(eachImgPath)
            rotateImgPath = rotator.rotateMain()

            if self.method == "bordered":
                bordered = BorderedTableDetector(rotateImgPath)
                bboxDtls, filename, cvImg = bordered.borderedMain()

            elif self.method == "unbordered":
                unbordered = UnborderedTableDetector(rotateImgPath)
                bboxDtls, filename, cvImg = unbordered.unborderedMain()

            elif self.method == "semibordered":
                semibordered = SemiborderedTableDetector(rotateImgPath)
                bboxDtls, filename, cvImg = semibordered.semiborderedMain()
            
            else:
                consolidated = ConsolidatedTableDetector(rotateImgPath)
                bboxDtls, filename, cvImg = consolidated.consolidatedMain()
            
        except:
            bboxDtls, filename= [], filename

        if len(bboxDtls) != 0:
            extracts = TableExtractor(ocrEngine, bboxDtls, cvImg)
            dataFrames = extracts.extractorMain()
        else:
            dataFrames = pd.DataFrame()

        return dataFrames

    def main(self):
        
        self.outImgFolder = './Images'

        if not os.path.exists(self.outImgFolder):
            os.makedirs(self.outImgFolder)
        
        reader = easyocr.Reader(['en'], gpu=False)

        try:
            if self.pdfPath.endswith('.PDF') or self.pdfPath.endswith('.pdf'):
                imagePath = PDFConverter.generateImage(self.pdfPath, self.outImgFolder)

            elif self.pdfPath.endswith('.jpg') or self.pdfPath.endswith('.JPG') or self.pdfPath.endswith('.png') or self.pdfPath.endswith('.PNG') or \
                self.pdfPath.endswith('.jpeg') or self.pdfPath.endswith('.JPEG'):

                img = cv2.imread(self.pdfPath)
                imgFolderPath = os.path.basename(self.pdfPath).replace('.jpg','').replace('.JPG','').replace('.jpeg','').replace('.JPEG','') \
                                            .replace('.png','').replace('.PNG','')
                if not os.path.exists(imgFolderPath):
                    os.makedirs(self.outImgFolder+"/"+imgFolderPath)
                
                cv2.imwrite(self.outImgFolder+"/"+imgFolderPath+"/"+os.path.basename(self.pdfPath), img)
                imagePath = [self.outImgFolder+"/"+imgFolderPath+"/"+os.path.basename(self.pdfPath)]
            
            else:
                print("Unsupported format")

        except Exception as e:
            print("Unsupported format")
        
        results = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(imagePath)) as executor:
            futures = [executor.submit(self.processImage, eachImgPath, reader) for eachImgPath in imagePath]
            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                results.append(result)
        try:
            shutil.rmtree(self.outImgFolder)
        except:
            pass

        return results


import cv2
import numpy as np
from scipy.ndimage import interpolation as inter

class ImageRotator:
    def __init__(self, imagePath):
        self.imagePath = imagePath

    def findScore(self, arr, angle):
        data = inter.rotate(arr, angle, reshape=False, order=0)
        hist = np.sum(data, axis=1)
        score = np.sum((hist[1:] - hist[:-1]) ** 2)
        return hist, score

    def findNormalRotation(self, img):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        ht, wd = img.shape[:2]
        pix = ~img
        binImg = 1 - (pix.reshape((ht, wd)) / 255.0)
        delta = 0.5
        limit = 6
        angles = np.arange(-limit, limit + delta, delta)
        scores = []
        for angle in angles:
            hist, score = self.findScore(binImg, angle)
            scores.append(score)

        bestScore = max(scores)
        bestAngle = angles[scores.index(bestScore)]
        return bestAngle

    def rotateImage(self, img, angle):
        h, w = img.shape[:2]
        center = (w // 2, h // 2)
        M = cv2.getRotationMatrix2D(center, angle, 1.0)
        rotated = cv2.warpAffine(img, M, (w, h),
                                 flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
        return rotated

    def alignImage(self):
        cvImage = cv2.imread(self.imagePath)
        res = cvImage.copy()
        normRotation = self.findNormalRotation(cvImage)
        rotatedImage = self.rotateImage(res, normRotation)
        return rotatedImage

    def rotateMain(self):
        correctedImage = self.alignImage()
        rotatePath = self.imagePath.replace('.jpg', '_rotated.jpg')
        cv2.imwrite(rotatePath, correctedImage)
        return rotatePath


import cv2
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None
import os

class BorderedTableDetector:
    def __init__(self, imgPath):
        self.imgPath = imgPath

    @staticmethod
    def sortContours(cnts, method='left-to-right'):

        reverse = False
        i = 0
        if method == "right-to-left" or method == "bottom-to-top":
            reverse = True
        if method == "top-to-bottom" or method == "bottom-to-top":
            i = 1
        boundingBoxes = [cv2.boundingRect(c) for c in cnts]
        (cnts, boundingBoxes) = zip(*sorted(zip(cnts, boundingBoxes),
                                            key=lambda b: b[1][i], reverse=reverse))
        
        return (cnts, boundingBoxes)

    def getProperTable(self, img):

        tableData = []
        grayImg = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        threshImg = cv2.threshold(grayImg, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]
        threshImg = 255 - threshImg
        kernel = np.ones((5,191), np.uint8)
        morphImg = cv2.morphologyEx(threshImg, cv2.MORPH_CLOSE, kernel)
        contours = cv2.findContours(morphImg, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        contours = contours[0] if len(contours) == 2 else contours[1]
        contours, boundingBoxes = self.sortContours(contours, method='top-to-bottom')
        res = img.copy()
        for c in contours:
            x, y, w, h = cv2.boundingRect(c)
            if (h>300):
                image = cv2.rectangle(res, (x,y), (x+w, y+h), (0,0,0), 12)
                tableData.append((x, y, w, h))
        
        return image, tableData

    @staticmethod
    def drawLines(img):
        
        res = img.copy()
        grayImg = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        threshImg = cv2.threshold(grayImg, 200, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

        vKernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 40))
        rmVertical = cv2.morphologyEx(threshImg, cv2.MORPH_OPEN, vKernel, iterations=2)

        hKernel = cv2.getStructuringElement(cv2.MORPH_RECT, (40, 1))
        rmHorizontal = cv2.morphologyEx(threshImg, cv2.MORPH_OPEN, hKernel, iterations=2)

        cnts = cv2.findContours(rmVertical, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        cnts = cnts[0] if len(cnts) == 2 else cnts[1]

        cnts1 = cv2.findContours(rmHorizontal, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        cnts1 = cnts1[0] if len(cnts1) == 2 else cnts1[1]

        for c in cnts:
            cv2.drawContours(res, [c], -1, (0,0,0), 4)

        for c1 in cnts1:
            cv2.drawContours(res, [c1], -1, (0,0,0), 4)
        
        return res

    @staticmethod
    def preProcess(img):
        
        grayImg = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        thresh, imgBin = cv2.threshold(grayImg, 128, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
        imgBin = 255 - imgBin

        return imgBin

    @staticmethod
    def verticalLineDetect(invertedImg):

        kernel_len = np.array(invertedImg).shape[1] // 100
        vKernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, kernel_len))
        vLines = cv2.dilate(cv2.erode(invertedImg, vKernel, iterations=3), vKernel, iterations=3)

        return vLines

    @staticmethod
    def horizonalLineDetect(invertedImg):
    
        kernel_len = np.array(invertedImg).shape[1] // 100
        hKernel = cv2.getStructuringElement(cv2.MORPH_RECT, (kernel_len, 1))
        hLines = cv2.dilate(cv2.erode(invertedImg, hKernel, iterations=3), hKernel, iterations=3)

        return hLines

    @staticmethod
    def combineLines(img, vLines, hLines):

        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (5, 5))
        imgVH = cv2.addWeighted(vLines, 0.5, hLines, 0.5, 0.0)
        thresh, imgVH = cv2.threshold(imgVH, 128, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
        grayImg = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        # bitNOT = cv2.bitwise_not(cv2.bitwise_xor(grayImg, imgVH))
        
        return imgVH

    def getBboxDtls(self, img, imgVH, tableDtls, imgPath):
        
        contours, hierarchy = cv2.findContours(imgVH, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        contours, boundingBoxes = self.sortContours(contours, method="top-to-bottom")

        boxes = []
        for c in contours:
            x, y, w, h = cv2.boundingRect(c)
            if (h>30) and (h<500):
                boxes.append([x, y, w, h])

        bboxDtls = {}
        for tableBox1 in tableDtls:
            key = tableBox1
            values = []

            for tableBox2 in boxes:
                x2, y2, w2, h2 = tableBox2
                if tableBox1[0] <= x2 <= tableBox1[0] + tableBox1[2] and tableBox1[1] <= y2 <= tableBox1[1] + tableBox1[3]:
                    values.append(tableBox2)
            bboxDtls[key] = values
        
        for key, values in bboxDtls.items():
            x_tab, y_tab, w_tab, h_tab = key
            cv2.rectangle(img, (x_tab, y_tab), (x_tab + w_tab, y_tab + h_tab), (255, 0, 0), 8)
            for box in values:
                x_box, y_box, w_box, h_box = box
                cv2.rectangle(img, (x_box, y_box), (x_box + w_box, y_box + h_box), (0, 255, 0), 4)

        cv2.imwrite(imgPath.replace('_rotated.jpg','_detected.jpg'),img)

        return bboxDtls

    def borderedMain(self):
        try:
            cvImg = cv2.imread(self.imgPath)
            fileName = os.path.basename(self.imgPath).replace('_rotated.jpg', '')
            tableImg, tableRect = self.getProperTable(cvImg)
            tableImg = self.drawLines(tableImg)
            res = tableImg.copy()
            preImg = self.preProcess(tableImg)
            vLineImg = self.verticalLineDetect(preImg)
            hLineImg = self.horizonalLineDetect(preImg)
            combineLineImg = self.combineLines(res, vLineImg, hLineImg)
            bboxDtls = self.getBboxDtls(tableImg, combineLineImg, tableRect, self.imgPath)
        except Exception as e:
            print(e)
            bboxDtls, fileName, res = [], fileName, None

        return bboxDtls, fileName, res


import cv2
import os
import numpy as np
import pandas as pd


class TableExtractor:
    def __init__(self, ocrEngine, tablebboxDtls, cvImg):
        self.ocrEngine = ocrEngine
        self.tablebboxDtls = tablebboxDtls
        self.cvImg = cvImg

    def getRowsCols(self, box, heights, mean):

        row = []
        column = []
        j = 0

        for i in range(len(box)):
            if (i == 0):
                column.append(box[i])
                previous = box[i]
            else:
                if (box[i][1] <= previous[1] + mean / 2):
                    column.append(box[i])
                    previous = box[i]
                    if (i == len(box) - 1):
                        row.append(column)
                else:
                    row.append(column)
                    column = []
                    previous = box[i]
                    column.append(box[i])

        countCol = 0
        index = 0
        for i in range(len(row)):
            current = len(row[i])
            if current > countCol:
                countCol = current
                index = i

        center = [int(row[i][j][0] + row[i][j][2] / 2) for j in range(len(row[i])) if row[0]]
        center = np.array(center)
        center.sort()

        finalBoxes = []
        for i in range(len(row)):
            lis = []
            for k in range(countCol):
                lis.append([])
            for j in range(len(row[i])):
                diff = abs(center - (row[i][j][0] + row[i][j][2] / 4))
                minimum = min(diff)
                indexing = list(diff).index(minimum)
                lis[indexing].append(row[i][j])
            finalBoxes.append(lis)
        
        return finalBoxes, row, countCol

    def extractBoxes(self, finalBoxes, img):

        outer = []
        for i in range(len(finalBoxes)):
            for j in range(len(finalBoxes[i])):
                inner = ''
                if (len(finalBoxes[i][j]) == 0):
                    outer.append(' ')
                else:
                    for k in range(len(finalBoxes[i][j])):
                        y, x, w, h = finalBoxes[i][j][k][0], finalBoxes[i][j][k][1], finalBoxes[i][j][k][2], finalBoxes[i][j][k][3]
                        finalImg = img[x:x + h, y:y + w]
                        _, imgCrop = cv2.threshold(cv2.cvtColor(finalImg, cv2.COLOR_BGR2GRAY), 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
                        out = self.ocrEngine.readtext(imgCrop, detail=0, paragraph=False)
                        for text in out:
                            inner = inner + text
                    outer.append(inner)
        
        return outer

    def dataframeGeneration(self, texts, rows, columnCount):

        arr = np.array(texts)
        dataFrame = pd.DataFrame(arr.reshape(len(rows), columnCount))
        data = dataFrame.style.set_properties(align='left')

        return data

    def extractorMain(self):#, tablebboxDtls, cvImg):

        allDataframes = []
        for tabs, boxes in self.tablebboxDtls.items():
            heights = [boxes[i][3] for i in range(len(boxes))]
            meanHeight = np.mean(heights)
            allSortedBoxes, rows, cols = self.getRowsCols(boxes, heights, meanHeight)
            extractedData = self.extractBoxes(allSortedBoxes, self.cvImg)
            eachDataDf = self.dataframeGeneration(extractedData, rows, cols)
            allDataframes.append(eachDataDf.data)

        return allDataframes


from io import open
from setuptools import setup, find_packages

with open('requirements.txt', encoding="utf-8-sig") as f:
    requirements = f.readlines()

def readme():
    with open('README.md', encoding="utf-8-sig") as f:
        README = f.read()
    return README

setup(
    name='tabledetector',
    packages=find_packages(),
    include_package_data=True,
    version='1.0.0',
    install_requires=requirements,
    license='MIT License',
    description='End-to-End table structure detector',
    long_description=readme(),
    long_description_content_type="text/markdown",
    author='Rishav Banerjee',
    author_email='rishavbanerjee10.rb@gmail.com',
    url='https://github.com/rajban94/TableDetector',
    download_url='https://github.com/rajban94/TableDetector.git',
    keywords=['table detector'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 1 - Planning',
        'Operating System :: OS Independent'
    ],

)